/*
 * Copyright (C) 2008 - Neil Kettle
 *
 * This file is part of equivset.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>

#include <cudd/cudd.h>
#include <cudd/cuddInt.h>

#include "antiunify.h"
#include "list.h"

#include "equiv.h"

static struct equiv_list *bdd_ComputeEquivalences (DdManager *,
                                                   struct equiv_cache *,
                                                   DdNode *);
struct equiv_list *
bddComputeEquivalences (DdManager *dd, DdNode *f)
{
  struct equiv_cache *cache;
  struct equiv_list *r, *rt;
  int i, nnodes;

  nnodes = Cudd_DagSize (f);

  cache = malloc (sizeof (struct equiv_cache));
  if (cache == NULL)
    {
      fprintf (stderr, "bddComputeEquivalences: failed allocating %d-bytes\n",
               sizeof (struct equiv_cache));
      exit (EXIT_FAILURE);
    }

  cache->table = st_init_table (st_ptrcmp, st_ptrhash);
  cache->idx = 0;
  cache->mx_idx = nnodes << 1;
  cache->empty = equiv_list_init_empty ();
  cache->page = malloc (sizeof (struct equiv_list *) * cache->mx_idx);
  if (cache->page == NULL)
    {
      fprintf (stderr, "bddComputeEquivalences: failed allocating %d-bytes\n",
               sizeof (struct equiv_list *) * cache->mx_idx);
      exit (EXIT_FAILURE);
    }

  rt = bdd_ComputeEquivalences (dd, cache, f);
  if (rt != CUDD_CONST_ONE && rt != CUDD_CONST_ZERO)
    r = equiv_list_copy (rt);
  else
    r = equiv_list_init_empty ();

  for (i = 0; i < cache->idx; i++)
    {
      if (cache->page[i] != cache->empty)
        equiv_list_free (cache->page[i]);
    }
  equiv_list_free (cache->empty);

  st_free_table (cache->table);
  free (cache->page);
  free (cache);

  return (r);
}

static struct equiv_list *
bdd_ComputeEquivalences (DdManager *dd, struct equiv_cache *cache, DdNode *f)
{
  struct equiv_list *n, **r, *rG, *rH;
  DdNode *F, *G, *g, *H, *h;

  F = Cudd_Regular (f);

  /* Terminal case, exit if constant */
  if (cuddIsConstant (F))
    return (struct equiv_list *) (f == dd->one ? CUDD_CONST_ONE
                                               : CUDD_CONST_ZERO);

  /* Terminal case, cache */
  if (F->ref != 1)
    if (st_lookup (cache->table, f, &r))
      return (*r);

  r = &cache->page[cache->idx++];

  g = cuddT (F);
  h = cuddE (F);
  if (Cudd_IsComplement (f))
    {
      g = Cudd_Not (g);
      h = Cudd_Not (h);
    }

  G = Cudd_Regular (g);
  H = Cudd_Regular (h);

  if (cuddIsConstant (G) || cuddIsConstant (H))
    {
      if (cuddIsConstant (G) && cuddIsConstant (H))
        {
          int pos, idx;

          rG = bdd_ComputeEquivalences (dd, cache, g);
          if (rG == NULL)
            return (NULL);

          pos = (rG == CUDD_CONST_ONE);
          idx = Cudd_ReadPerm (dd, F->index);

#ifdef DEBUG
          printf ("%d: both constants, returning\n", idx);
#endif

          n = equiv_list_init_single (idx, pos, !pos);
        }
      else if (cuddIsConstant (G))
        {
#ifdef DEBUG
          printf ("%d: t constant, doing e: %p\n",
                  Cudd_ReadPerm (dd, F->index), h);
#endif

          rG = bdd_ComputeEquivalences (dd, cache, g);
          if (rG == NULL)
            return (NULL);

          if (rG != CUDD_CONST_ONE)
            {
              int idx = Cudd_ReadPerm (dd, F->index);

              rH = bdd_ComputeEquivalences (dd, cache, h);
              if (rH == NULL)
                return (NULL);

              n = equiv_list_copy_insert (rH, idx, 0, 1);
            }
          else
            {
              n = cache->empty;
            }
        }
      else
        {
#ifdef DEBUG
          printf ("%d: e constant, doing t: %p\n",
                  Cudd_ReadPerm (dd, F->index), g);
#endif

          rH = bdd_ComputeEquivalences (dd, cache, h);
          if (rH == NULL)
            return (NULL);

          if (rH != CUDD_CONST_ONE)
            {
              int idx = Cudd_ReadPerm (dd, F->index);

              rG = bdd_ComputeEquivalences (dd, cache, g);
              if (rG == NULL)
                return (NULL);

              n = equiv_list_copy_insert (rG, idx, 1, 0);
            }
          else
            {
              n = cache->empty;
            }
        }
    }
  else
    {
      rG = bdd_ComputeEquivalences (dd, cache, g);
      if (rG == NULL)
        return (NULL);

      if (LIST_NOT_EMPTY (rG))
        {
          rH = bdd_ComputeEquivalences (dd, cache, h);
          if (rH == NULL)
            return (NULL);

          if (LIST_NOT_EMPTY (rH))
            {
              struct equiv_list *rG_list, *rH_list, *rt;
              int idx = Cudd_ReadPerm (dd, F->index);

              rH_list = equiv_list_copy_insert (rH, idx, 0, 1);
              rG_list = equiv_list_copy_insert (rG, idx, 1, 0);

#ifdef DEBUG
              printf ("antiunify - input\n");
              equiv_list_dump (rH_list);
              equiv_list_dump (rG_list);
#endif

              rt = anti_unify_equiv (rH_list, rG_list);
              if (rt == NULL)
                n = cache->empty;
              else
                {
                  n = equiv_list_copy (rt);
                  equiv_list_free (rt);
                }

#ifdef DEBUG
              printf ("antiunify - output\n");
              equiv_list_dump (n);
#endif

              equiv_list_free (rH_list);
              equiv_list_free (rG_list);
            }
          else
            {
              n = cache->empty;
            }
        }
      else
        {
          n = cache->empty;
        }
    }

  *r = n;

  if (F->ref != 1)
    if (st_insert (cache->table, (char *) f, (char *) r) == ST_OUT_OF_MEM)
      return (NULL);

  return (*r);
}
